/*
  Doc_Arduino - german Arduino Forum
  IDE 1.8.13
  Arduino Nano Every
  18.03.2020
  TCBn 16Bit Timer (megaAVR0 TCBn Timer has no OVF Flag)
  License: GNU GPLv3
*/
/*
 * Hiermit kann man erste Versuche mit der Frequenz- und Pulsweitenmessung machen.
 * Dazu verbindet man Pin 13 (Taktpin) mit Pin 21 (TCB Messpin). 
 * Pin 13 wird getoggelt wenn die loop gerade Zeit hat.
 * Entsprechend "zappelt" die gemessene Periodendauer.
 */
 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <Streaming.h>
#include <NanoEveryPin.h>
#include <megaAVR0TCBn.h>

OutputPin <13> onboardLed;    // PE2
InputPin  <21> signalEingang; // PD5

TCB <0> timerTCB0;
/* in Funktion ab Zeile 90
   PRESCALER  1 wenn timerTCB0.setCLKDIV1() 
   PRESCALER  2 wenn timerTCB0.setCLKDIV2() 
   PRESCALER 64 wenn timerTCB0.setCLKTCA()  */
const unsigned int PRESCALER {64};

// --------------------------------------------------------

void setup (void)
{
  Serial.begin(250000);
  Serial.println("\nSTART #### #### #### ####");
  initPins();
  initEventSystem();
  initTCB0();
}

void loop (void)
{ 
  // TCB0 Messbereich beachten
  heartbeat(250); // Periodendauer 

  serialMonitor(1000);
}

// ****** Funktionen ****************************** //

void serialMonitor (const uint32_t interval)
{
  static uint32_t lastMillis {0};

  if (millis() - lastMillis > interval)
  {
    lastMillis += interval;

    uint16_t periode {0};
    uint16_t puls {0};
    ATOMIC_BLOCK (ATOMIC_RESTORESTATE)
    {
      periode = timerTCB0.cnt;
      puls = timerTCB0.ccmp;
    }

    Serial << "T " << periode << " \t" << "P " << puls << " \t" <<
           (62.5*PRESCALER*periode/1000000UL) << "ms \t" << (62.5*PRESCALER*puls/1000000UL) << "ms" <<
           // Verwendeten Prescaler beachten! //
    endl;
  }
}

void initPins (void)
{
  signalEingang.init();
  signalEingang.enablePullup();
  onboardLed.init();
}

void initEventSystem (void)
{
  EVSYS.CHANNEL2 = EVSYS_GENERATOR_PORT1_PIN5_gc;   // connect Pin 'PD5' to channel 2
  EVSYS.USERTCB0 = EVSYS_CHANNEL_CHANNEL2_gc;       // connect 'generator 'PD5' with user 'TCB0' via channel 2
}

void initTCB0 (void)
{
  timerTCB0.reset();
  //timerTCB0.setCLKDIV1();         // Messung einer Periodendauer bis 4,09ms möglich
  //timerTCB0.setCLKDIV2();         // Messung einer Periodendauer bis 8,1ms möglich
  timerTCB0.setCLKTCA();            // Arduino TCA0 default Prescaler 64, Periodendauer bis 262,1ms möglich
  timerTCB0.setModeFRQPW();
  timerTCB0.enableEventInput();
  timerTCB0.enableCAPTInterrupt();
  timerTCB0.enable();
}

ISR(TCB0_INT_vect)
{
  timerTCB0.cnt = TCB0.CNT;         // read out CNT first, then CCMP
  timerTCB0.ccmp = TCB0.CCMP;       // the CAPT flag is automatically clear after reading CCMP
}

void heartbeat (const unsigned long periode)
{
  const unsigned long interval {periode/2};
  static unsigned long lastMillis {0};

  if (millis() - lastMillis > interval)
  {
    lastMillis += interval;
    onboardLed.toggle();
  }
}
