/*
  Doc_Arduino - german Arduino Forum
  IDE 1.8.13
  Arduino Nano Every
  18.03.2020
  TCA0 16Bit Timer
  TCBn 16Bit Timer (megaAVR0 TCBn Timer has no OVF Flag)
  License: GNU GPLv3
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <Streaming.h>
#include <NanoEveryPin.h>
#include <megaAVR0TCBn.h>

OutputPin <13> onboardLed;    // PE2
InputPin  <21> signalEingang; // PD5

TCB <0> timerTCB0;
// PRESCALER 1 wenn timerTCB0.setCLKDIV1() //
// PRESCALER 2 wenn timerTCB0.setCLKDIV2() //
// PRESCALER ? wenn timerTCB0.setCLKTCA()  //
const unsigned int PRESCALER {16};
const unsigned int PERIODE {11000};        // 16Bit Wertebereich

// --------------------------------------------------------

void setup (void)
{
  Serial.begin(250000);
  Serial.println("\nSTART #### #### #### ####");
  initPins();
  initEventSystem();
  initTCB0();
  
  if (configTCA0(PRESCALER, PERIODE)) {
    runTCA0();
  }
  else {
    Serial.println(F("ungültiger TCA0 Prescaler"));
  }

  Serial << "Soll Periodendauer " << 62.5*PRESCALER*PERIODE/1000000UL << "ms" << endl;
}

void loop (void)
{
  serialMonitor(1000);
}

// ****** Funktionen ****************************** //

unsigned long correctedMillis (const unsigned int prescaler)
{
  const unsigned long long temp = 1ULL * millis() * prescaler / 64; // TCA0 Prescaler 64 ist Arduino Default
  return static_cast<unsigned long>(temp);
}

void serialMonitor (const uint32_t interval)
{
  static uint32_t lastMillis {0};

  //if (millis() - lastMillis > interval)
  if (correctedMillis(PRESCALER) - lastMillis > interval)
  {
    lastMillis += interval;

    uint16_t periode {0};
    uint16_t puls {0};
    ATOMIC_BLOCK (ATOMIC_RESTORESTATE)
    {
      periode = timerTCB0.cnt;
      puls = timerTCB0.ccmp;
    }

    Serial << "T " << periode << "  " << "P " << puls << "   " <<
           (62.5*PRESCALER*periode/1000000UL) << "ms  " << (62.5*PRESCALER*puls/1000000UL) << "ms" <<
           // Verwendeten Prescaler beachten! //
    endl;
  }
}

void initPins (void)
{
  signalEingang.init();
  signalEingang.enablePullup();
  onboardLed.init();
}

void initEventSystem (void)
{
  EVSYS.CHANNEL2 = EVSYS_GENERATOR_PORT1_PIN5_gc;   // connect Pin 'PD5' to channel 2
  EVSYS.USERTCB0 = EVSYS_CHANNEL_CHANNEL2_gc;       // connect 'generator 'PD5' with user 'TCB0' via channel 2
}

void initTCB0 (void)
{
  timerTCB0.reset();
  //timerTCB0.setCLKDIV1();         // Messung einer Periodendauer bis 4,09ms möglich
  //timerTCB0.setCLKDIV2();         // Messung einer Periodendauer bis 8,1ms möglich
  timerTCB0.setCLKTCA();            // Arduino default 64, Messung einer Periodendauer bis 262,1ms möglich
  timerTCB0.setModeFRQPW();
  timerTCB0.enableEventInput();
  timerTCB0.enableCAPTInterrupt();
  timerTCB0.enable();
}

ISR(TCB0_INT_vect)
{
  timerTCB0.cnt = TCB0.CNT;         // read out CNT first, then CCMP
  timerTCB0.ccmp = TCB0.CCMP;       // the CAPT flag is automatically clear after reading CCMP
}

// -------------------------------------------------------------------------------

ISR(TCA0_OVF_vect)
{
  onboardLed.toggle();
  TCA0.SINGLE.INTFLAGS = TCA0.SINGLE.INTFLAGS | TCA_SINGLE_OVF_bm;  // delete Interrupt Flag
}

bool configTCA0 (const unsigned int prescaler, const unsigned int top)
{
  bool state = true;

  TCA0.SINGLE.CTRLA = 0;  // stop/reset
  TCA0.SINGLE.CTRLB = 0;
  TCA0.SINGLE.CTRLC = 0;
  TCA0.SINGLE.CTRLD = 0;
  TCA0.SINGLE.EVCTRL = 0;
  PORTMUX.TCAROUTEA = 0;
  TCA0.SINGLE.CTRLB = TCA0.SINGLE.CTRLB | TCA_SINGLE_WGMODE_NORMAL_gc;  
  TCA0.SINGLE.INTCTRL = TCA0.SINGLE.INTCTRL | TCA_SINGLE_OVF_bm;       
  TCA0.SINGLE.CNT = 0;
  TCA0.SINGLE.PERBUF = top/2;
  
  ATOMIC_BLOCK (ATOMIC_RESTORESTATE)
  {
    switch (prescaler)  // set Prescaler Clock Select Bits
    {
      case    1: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV1_gc;     break;
      case    2: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV2_gc;     break;
      case    4: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV4_gc;     break;
      case    8: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV8_gc;     break;
      case   16: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV16_gc;    break;
      case   64: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV64_gc;    break;  
      case  256: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV256_gc;   break;
      case 1024: TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV1024_gc;  break;
      default  : state = false;  break;
    }
  }
  return state;
}

void runTCA0 (void)
{
  TCA0.SINGLE.CTRLA = TCA0.SINGLE.CTRLA | TCA_SINGLE_ENABLE_bm;       // start TCA0 timer
}
